package com.xpo.ltl.api.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.commons.fileupload.MultipartStream;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpEntityEnclosingRequestBase;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.HttpClientUtils;
import org.apache.http.conn.ConnectionPoolTimeoutException;
import org.apache.http.entity.BufferedHttpEntity;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.entity.mime.FormBodyPartBuilder;
import org.apache.http.entity.mime.HttpMultipartMode;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.pool.PoolStats;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParseException;
import com.google.gson.JsonParser;
import com.google.gson.reflect.TypeToken;
import com.xpo.ltl.api.client.common.Attachment;
import com.xpo.ltl.api.client.common.ClientCredentials;
import com.xpo.ltl.api.client.common.HyperMedia;
import com.xpo.ltl.api.client.common.Response;
import com.xpo.ltl.api.client.exception.ConflictException;
import com.xpo.ltl.api.client.exception.FaultResponse;
import com.xpo.ltl.api.client.exception.MoreInfo;
import com.xpo.ltl.api.client.exception.NotFoundException;
import com.xpo.ltl.api.client.exception.NotModifiedException;
import com.xpo.ltl.api.client.exception.ServiceException;
import com.xpo.ltl.api.client.exception.ValidationException;
import com.xpo.ltl.api.client.util.DateTimeHelper;

public class Resource<T> {

	private static final Log LOG = LogFactory.getLog(Resource.class);

	private static final String PROPERTY_FILE = "xpo-sdk-utils";

	private static final String PROPERTY_MAX_CONNECTIONS = "MAX_CONNECTIONS";
	private static final int DEFAULT_MAX_CONNECTIONS = 100;

	private static final String PROPERTY_MAX_CONNECTIONS_PER_ROUTE = "MAX_CONNECTIONS_PER_ROUTE";
	private static final int DEFAULT_MAX_CONNECTIONS_PER_ROUTE = 100;

	private static final String PROPERTY_CONNECTION_REQUEST_TIMEOUT = "CONNECTION_REQUEST_TIMEOUT";
	private static final int DEFAULT_CONNECTION_REQUEST_TIMEOUT = 1000; // 1 second default

	private static final String PROPERTY_CONNECTION_TIMEOUT = "CONNECTION_TIMEOUT";
	private static final int DEFAULT_CONNECTION_TIMEOUT = 1000; // 1 second default

	private static final String PROPERTY_SOCKET_TIMEOUT = "SOCKET_TIMEOUT";
	private static final int DEFAULT_SOCKET_TIMEOUT = 120000; // 2 minute default

	private static final Pattern FILENAME_PATTERN = Pattern.compile(".*filename=\"(.*)\".*", Pattern.DOTALL);
	private static final Pattern CONTENT_TYPE_PATTERN = Pattern.compile(".*Content-Type: (.*)", Pattern.DOTALL);

    private static PoolingHttpClientConnectionManager cm = null;

	private static final Gson DEFAULT_GSON = new GsonBuilder()
			.registerTypeHierarchyAdapter(XMLGregorianCalendar.class, new XMLGregorianCalendarAdapter())
			.setFieldNamingStrategy(new XmlElementFieldNamingStrategy())
			.create();
	private static final CloseableHttpClient client;
	private String host;
	private String basePath;
	private String resource;
	private String accessToken;
	private Gson gson;
	private final HashMap<String, List<String>> parameters;
	private final MultipartEntityBuilder mpEntityBuilder;
	private final List<FormBodyPartBuilder> formBuilder;
	private final String boundary;
	private final Class<T> type;

	static {
		LOG.info("Creating PoolingHttpClientConnectionManager");
		cm = new PoolingHttpClientConnectionManager();

		final int maxConnections = getIntProperty(PROPERTY_MAX_CONNECTIONS, DEFAULT_MAX_CONNECTIONS);
		LOG.info("PoolingHttpClientConnectionManager: setting MaxTotal to: " + maxConnections);
		cm.setMaxTotal(maxConnections);

		final int maxConnectionsPerRoute = getIntProperty(PROPERTY_MAX_CONNECTIONS_PER_ROUTE,
				DEFAULT_MAX_CONNECTIONS_PER_ROUTE);
		LOG.info("PoolingHttpClientConnectionManager: setting DefaultMaxPerRoute to: " + maxConnectionsPerRoute);
		cm.setDefaultMaxPerRoute(maxConnectionsPerRoute);

		final int connectionRequestTimeout = getIntProperty(PROPERTY_CONNECTION_REQUEST_TIMEOUT,
				DEFAULT_CONNECTION_REQUEST_TIMEOUT);
		final int connectionTimeout = getIntProperty(PROPERTY_CONNECTION_TIMEOUT, DEFAULT_CONNECTION_TIMEOUT);
		final int socketTimeout = getIntProperty(PROPERTY_SOCKET_TIMEOUT, DEFAULT_SOCKET_TIMEOUT);

		LOG.info("Building RequestConfig with ConnectionRequestTimeout: " + connectionRequestTimeout
				+ ", ConnectionTimeout: " + connectionTimeout + ", SocketTimeout: " + socketTimeout);
		final RequestConfig config = RequestConfig.custom().setConnectionRequestTimeout(connectionRequestTimeout)
				.setConnectTimeout(connectionTimeout).setSocketTimeout(socketTimeout).setExpectContinueEnabled(true)
				.build();

		LOG.info("Building HttpClient");
		client = HttpClients.custom().setConnectionManager(cm).setDefaultRequestConfig(config).build();
	}

	public Resource(final String host, final String basePath, final String resource, final String accessToken, final Class<T> type) {
		this(host, basePath, resource, accessToken, type, null);
	}
	
	public Resource(final String host, final String basePath, final String resource, final String accessToken, final Class<T> type, Gson gson) {
		super();
		this.host = host;
		this.basePath = basePath;
		this.resource = resource;
		this.setAccessToken(accessToken);
		this.type = type;
		parameters = new HashMap<String, List<String>>();
		formBuilder = new ArrayList<FormBodyPartBuilder>();
		mpEntityBuilder = MultipartEntityBuilder.create();
		boundary = "----" + System.currentTimeMillis();
		mpEntityBuilder.setBoundary(boundary);
		mpEntityBuilder.setMode(HttpMultipartMode.RFC6532);
		if(gson != null) {
			this.gson = gson;
		} else {
			this.gson = DEFAULT_GSON;
		}
	}

	protected T getToken(final ClientCredentials credentials) throws ServiceException {
		CloseableHttpResponse r = null;
		try {
			String authKey = credentials.getAuthKey();
			String user = credentials.getUsername();
			String password = credentials.getPassword();
			String scope = credentials.getScope();
			final HttpPost hPost = new HttpPost(getURL());
			hPost.addHeader("Authorization", "Basic " + authKey);
			hPost.addHeader("Content-Type", "application/x-www-form-urlencoded");
			if(scope == null) {
				hPost.setEntity(new StringEntity("grant_type=password&username=" + user + "&password=" + password));
			} else {
				hPost.setEntity(new StringEntity("grant_type=password&username=" + user + "&password=" + password + "&scope=" + scope));
			}
			checkAndLogStats();
			r = client.execute(hPost);
			final BufferedHttpEntity bhe = new BufferedHttpEntity(r.getEntity());
			final byte[] buf = new byte[(int) bhe.getContentLength()];
			bhe.getContent().read(buf);
			if (r.getStatusLine().getStatusCode() != 200) {
				throw new Exception(new String(buf));
			} else {
				return gson.fromJson(new String(buf), type);
			}
		} catch (final ConnectionPoolTimeoutException e){
			LOG.error("Resource getToken(String, String, String) ConnectionPoolTimeoutException! ", e);
			throw new ServiceException("90000", e.toString());
		} catch (final Exception e) {
			LOG.error("Resource getToken(String, String, String) Exception! ", e);
			throw new ServiceException("90000", e.toString());
		} finally {
			HttpClientUtils.closeQuietly(r);
		}
	}

	protected T refreshToken(final String refreshToken) throws ServiceException {
		CloseableHttpResponse r = null;
		try {
			final HttpPost hPost = new HttpPost(getURL());
			hPost.addHeader("Content-Type", "application/x-www-form-urlencoded");
			hPost.setEntity(new StringEntity("grant_type=refresh_token&refresh_token=" + refreshToken));
			checkAndLogStats();
			r = client.execute(hPost);
			final BufferedHttpEntity bhe = new BufferedHttpEntity(r.getEntity());
			final byte[] buf = new byte[(int) bhe.getContentLength()];
			bhe.getContent().read(buf);
			if (r.getStatusLine().getStatusCode() != 200) {
				throw new Exception(new String(buf));
			} else {
				return gson.fromJson(new String(buf), type);
			}
		} catch (final ConnectionPoolTimeoutException e){
			LOG.error("Resource refreshToken(String) ConnectionPoolTimeoutException! ", e);
			throw new ServiceException("90000", e.toString());
		} catch (final Exception e) {
			LOG.error("Resource refreshToken(String) Exception! ", e);
			throw new ServiceException("90000", e.toString());
		} finally {
			HttpClientUtils.closeQuietly(r);
		}
	}

	public Response<T> post(final Object payload) throws ServiceException {
		try {
			final HttpPost hPost = new HttpPost(getURL());
			hPost.setEntity(new StringEntity(gson.toJson(payload)));
			hPost.addHeader("Content-Type", "application/json");
			return execute(hPost);
		} catch (final ServiceException e) {
			LOG.error("Resource post(Object) Exception! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource post(Object) Exception! ", e);
			throw new ServiceException("90000", e.toString());
		}
	}

	public Response<T> post() throws ServiceException {
		try {
			final HttpPost hPost = new HttpPost(getURL());
			hPost.addHeader("Content-Type", "application/json");
			return execute(hPost);
		} catch (final ServiceException e) {
			LOG.error("Resource post() Exception! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource post() Exception! ", e);
			throw new ServiceException("90000", e.toString());
		}
	}

	public Response<T> postForm() throws ServiceException {
		try {
			final HttpPost hPost = new HttpPost(getURL());
			buildForm(hPost);
			return execute(hPost);
		} catch (final ServiceException e) {
			LOG.error("Resource postForm() Exception! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource postForm() Exception! ", e);
			throw new ServiceException("90000", e.toString());
		}
	}

	private void buildForm(final HttpEntityEnclosingRequestBase hPost) {
		hPost.addHeader("Content-Type", "multipart/form-data; boundary=" + boundary);
		if(!formBuilder.isEmpty()) {
			for(final FormBodyPartBuilder fb: formBuilder) {
				mpEntityBuilder.addPart(fb.build());
			}
		}
		hPost.setEntity(mpEntityBuilder.build());
	}

	public Response<T> get() throws ServiceException {
		final HttpGet hGet = new HttpGet(getURL());
		return execute(hGet);
	}

	public Response<T> delete(final Long versionCheckTsInMillis) throws ServiceException {
		final HttpDelete hDelete = new HttpDelete(getURL());
		hDelete.addHeader("Transaction-Timestamp", versionCheckTsInMillis.toString());
		return execute(hDelete);
	}

	public Response<T> put(final Object payload, final Long versionCheckTsInMillis) throws ServiceException {
		try {
			final HttpPut hPut = new HttpPut(getURL());
			if(payload != null) {
				hPut.setEntity(new StringEntity(gson.toJson(payload)));
			}
			hPut.addHeader("Content-Type", "application/json");
			hPut.addHeader("Transaction-Timestamp", versionCheckTsInMillis.toString());
			return execute(hPut);
		} catch (final ServiceException e) {
			LOG.error("Resource put(Object, Long) Exception! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource put(Object, Long) Exception! ", e);
			throw new ServiceException("90000", e.toString());
		}
	}

	public Response<T> put(final Long versionCheckTsInMillis) throws ServiceException {
		return put(null, versionCheckTsInMillis);
	}

	public Response<T> putForm() throws ServiceException {
		try {
			final HttpPut hPut = new HttpPut(getURL());
			buildForm(hPut);
			return execute(hPut);
		} catch (final ServiceException e) {
			LOG.error("Resource putForm() Exception! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource putForm() Exception! ", e);
			throw new ServiceException("90000", e.toString());
		}
	}

	private String getURL() {
		final StringBuilder sb = new StringBuilder(host);
		if (basePath != null && basePath.length() > 0) {
			if (!basePath.startsWith("/")) {
				sb.append("/");
			}
			sb.append(basePath);
		}
		if (resource != null && resource.length() > 0) {
			if (!resource.startsWith("/")) {
				sb.append("/");
			}
			sb.append(resource);
		}
		if (parameters.size() > 0) {
			sb.append("?");
			boolean firstParam = true;
			for (final String p : parameters.keySet()) {
				if (!firstParam) {
					sb.append("&");
				}
				firstParam = false;
				boolean firstValue = true;
				for (final String v : parameters.get(p)) {
					if (!firstValue) {
						sb.append("&");
					}
					firstValue = false;
					sb.append(p);
					sb.append("=");
					try {
						sb.append(URLEncoder.encode(v, "UTF-8"));
					} catch (final UnsupportedEncodingException e) {
						LOG.warn("Resource getURL() UnsupportedEncodingException! ", e);
						sb.append(v);
					}
				}
			}
		}
		return sb.toString();
	}

	private Response<T> execute(final HttpRequestBase rqst) throws ServiceException {
		CloseableHttpResponse r = null;
		try {
			if (accessToken.toLowerCase().startsWith("bearer")) {
				rqst.addHeader("Authorization", accessToken);
			} else {
				rqst.addHeader("Authorization", "Bearer " + accessToken);
			}
			checkAndLogStats();
			r = client.execute(rqst);
			return handleEntity(r);
		} catch (final ConnectionPoolTimeoutException e){
			LOG.error("Resource execute(HttpRequestBase) ConnectionPoolTimeoutException! ", e);
			throw new ServiceException("90000", e.toString());
		} catch (final ServiceException e) {
			LOG.error("Resource execute(HttpRequestBase) ServiceException! ", e);
			throw e;
		} catch (final Exception e) {
			LOG.error("Resource execute(HttpRequestBase) Exception! ", e);
			throw new ServiceException("90000", e.toString());
		} finally {
			HttpClientUtils.closeQuietly(r);
		}
	}

	private Response<T> handleEntity(final HttpResponse httpResp) throws ServiceException {
		final HttpEntity entity = httpResp.getEntity();
		Response<T> response = new Response<T>();
		final int statusCode = httpResp.getStatusLine().getStatusCode();
		response.setCode(Integer.toString(statusCode));
		if (entity == null) {
			return response;
		}
		try {
			final ContentType ct = ContentType.get(entity);
			if (ct.getMimeType().equals("multipart/form-data")) {
				final List<Attachment> attachment = new ArrayList<Attachment>();
				final String boundary = ct.getParameter("boundary");
				final MultipartStream ms = new MultipartStream(entity.getContent(), boundary.getBytes(), 4096, null);
				boolean nextPart = ms.skipPreamble();
				boolean firstPart = true;
				while (nextPart) {
					String headers = ms.readHeaders();
					final ByteArrayOutputStream os = new ByteArrayOutputStream();
					ms.readBodyData(os);
					if (firstPart) {
						response = parseJson(os.toString());
						firstPart = false;
					} else {
						final Attachment f = new Attachment();
						if (headers != null) {
							headers = headers.trim();
							Matcher m1 = FILENAME_PATTERN.matcher(headers);
							if (m1.matches()) {
								f.setFileName(m1.group(1));
							}
							m1 = CONTENT_TYPE_PATTERN.matcher(headers);
							if (m1.matches()) {
								f.setContentType(m1.group(1));
							}
						}
						f.setContents(os.toByteArray());
						attachment.add(f);
					}
					nextPart = ms.readBoundary();
				}
				response.setAttachment(attachment);
				return response;
			} else {
				final BufferedHttpEntity bhe = new BufferedHttpEntity(entity);
				final byte[] buf = new byte[(int) bhe.getContentLength()];
				bhe.getContent().read(buf);
				if (ct.getMimeType().equals("application/json")) {
					response = parseJson(new String(buf));
				} else {
					final FaultResponse fault = new FaultResponse();
					fault.setMessage(new String(buf));
					fault.setErrorCode("500");
					throw new ServiceException(fault);
				}
				return response;
			}
		} catch (final IOException e) {
			LOG.error("Resource handleEntity(HttpResponse) IOException! ", e);
			throw new ServiceException("90000", e.getMessage());
		}
	}

	@SuppressWarnings("unchecked")
	private Response<T> parseJson(final String json) throws ServiceException {
		final Response<T> response = new Response<T>();
		final JsonParser parser = new JsonParser();
		Object obj = parser.parse(json);
		JsonObject o = null;
		if(obj instanceof JsonObject) {
			o = (JsonObject) obj;
		} else {
			JsonArray jsa = (JsonArray)obj;
			List<T> list = new ArrayList<T>();
			for(JsonElement e: jsa) {
				list.add(gson.fromJson(e, type));
			}
			response.setListData(list);
			return response;
		}
		if(o.get("code") == null) {
			response.setData(gson.fromJson(o, type));
			return response;
		}
		response.setCode(o.get("code").getAsString());
		try {
			final JsonElement tranTmst = o.get("transactionTimestamp");
			if(tranTmst != null) {
				try {
					response.setTransactionTimestamp(DateTimeHelper.convertTimeInMillisToXMLGC(tranTmst.getAsLong()));
				} catch(NumberFormatException e) {
					response.setTransactionTimestamp(DateTimeHelper.convertStringToXMLGC(tranTmst.getAsString()));
				}
			}
		} catch (final Exception e) {
			LOG.warn("Resource parseJson(String) DatatypeConfigurationException! ", e);
		}
		final JsonElement data = o.get("data");
		if (response.getCode().startsWith("2")) {
			if (data != null) {
				response.setData(gson.fromJson(data, type));
			}
			final JsonElement link = o.get("link");
			if (link != null) {
				response.setLink((List<HyperMedia>) gson.fromJson(link, new TypeToken<List<HyperMedia>>() {
				}.getType()));
			}
		} else {
			final JsonElement error = o.get("error");
			if (error != null) {
				try {
					final FaultResponse fault = (gson.fromJson(error, FaultResponse.class));
					if(response.getCode().equals("400")) {
						throw new ValidationException(fault);
					}
					if(response.getCode().equals("404")) {
						throw new NotFoundException(fault);
					}
					if(response.getCode().equals("409")) {
						throw new ConflictException(fault);
					}
					if(response.getCode().equals("304")) {
						throw new NotModifiedException(fault);
					}
					throw new ServiceException(fault);
				} catch(JsonParseException e) {
					LOG.error("Resource parseJson(String) Parsing Error! error: " +  error);
					FaultResponse fault = new FaultResponse();
					fault.setErrorCode("500");
					fault.setMessage(e.getMessage());
					MoreInfo info = new MoreInfo();
					info.setLocation("JSON Payload");
					info.setMessage(error.toString());
					fault.getMoreInfo().add(info);
					throw new ServiceException(fault);
				}
			}
		}
		return response;
	}

	public String getHost() {
		return host;
	}

	public void setHost(final String host) {
		this.host = host;
	}

	protected String getBasePath() {
		return basePath;
	}

	protected void setBasePath(final String basePath) {
		this.basePath = basePath;
	}

	public String getResource() {
		return resource;
	}

	public void setResource(final String resource) {
		this.resource = resource;
	}

	public String getAccessToken() {
		return accessToken;
	}

	public void setAccessToken(final String accessToken) {
		this.accessToken = accessToken;
	}

	public HashMap<String, List<String>> getParameters() {
		return parameters;
	}

	public void addParameter(final String name, final Double value) {
		if (value == null) {
			return;
		}
		addParameter(name, value.toString());
	}

	public void addParameter(final String name, final Double[] value) {
		if (value == null) {
			return;
		}
		for (final Double v: value) {
			addParameter(name, v.toString());
		}
	}

	public void addParameter(final String name, final Integer value) {
		if (value == null) {
			return;
		}
		addParameter(name, value.toString());
	}

	public void addParameter(final String name, final Integer[] value) {
		if (value == null) {
			return;
		}
		for (final Integer v: value) {
			addParameter(name, v.toString());
		}
	}

	public void addParameter(final String name, final Long value) {
		if (value == null) {
			return;
		}
		addParameter(name, value.toString());
	}

	public void addParameter(final String name, final Long[] value) {
		if (value == null) {
			return;
		}
		for (final Long v: value) {
			addParameter(name, v.toString());
		}
	}

	public void addParameter(final String name, final Boolean value) {
		if (value == null) {
			return;
		}
		addParameter(name, value.toString());
	}

	public void addParameter(final String name, final Boolean[] value) {
		if (value == null) {
			return;
		}
		for (final Boolean v: value) {
			addParameter(name, v.toString());
		}
	}

	public void addParameter(final String name, final Byte value) {
		if (value == null) {
			return;
		}
		addParameter(name, value.toString());
	}

	public void addParameter(final String name, final Byte[] value) {
		if (value == null) {
			return;
		}
		for (final Byte v: value) {
			addParameter(name, v.toString());
		}
	}

	public void addParameter(final String name, final XMLGregorianCalendar value) {
		if (value == null) {
			return;
		}
		addParameter(name, Long.toString(value.toGregorianCalendar().getTimeInMillis()));
	}

	public void addParameter(final String name, final XMLGregorianCalendar[] value) {
		if (value == null) {
			return;
		}
		for (final XMLGregorianCalendar v: value) {
			addParameter(name, Long.toString(v.toGregorianCalendar().getTimeInMillis()));
		}
	}

	public void addParameter(final String name, final String value) {
		if (value == null) {
			return;
		}
		final List<String> values = getValues(name);
		values.add(value);
	}

	public void addParameter(final String name, final String[] value) {
		if (value == null) {
			return;
		}
		final List<String> values = getValues(name);
		for (final String v: value) {
			values.add(v);
		}
	}

	private List<String> getValues(final String name) {
		List<String> values = parameters.get(name);
		if (values == null) {
			values = new ArrayList<String>();
			parameters.put(name, values);
		}
		return values;
	}

	public void addFormData(final String name, final Double value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final Double[] value) {
		if (value == null) {
			return;
		}
		for (final Double v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final Integer value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final Integer[] value) {
		if (value == null) {
			return;
		}
		for (final Integer v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final Long value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final Long[] value) {
		if (value == null) {
			return;
		}
		for (final Long v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final Boolean value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final Boolean[] value) {
		if (value == null) {
			return;
		}
		for (final Boolean v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final Byte value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final Byte[] value) {
		if (value == null) {
			return;
		}
		for (final Byte v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final Attachment value) {
		if (value == null) {
			return;
		}
		mpEntityBuilder.addBinaryBody(name, value.getContents(), ContentType.create(value.getContentType()), value.getFileName());
	}

	public void addFormData(final String name, final Attachment[] value) {
		if (value == null) {
			return;
		}
		for (final Attachment v: value) {
			mpEntityBuilder.addBinaryBody(name, v.getContents(), ContentType.create(v.getContentType()), v.getFileName());
		}
	}

	public void addFormData(final String name, final XMLGregorianCalendar value) {
		if (value == null) {
			return;
		}
		addFormData(name, value.toString());
	}

	public void addFormData(final String name, final XMLGregorianCalendar[] value) {
		if (value == null) {
			return;
		}
		for (final XMLGregorianCalendar v: value) {
			addFormData(name, v.toString());
		}
	}

	public void addFormData(final String name, final String value) {
		if (value == null) {
			return;
		}
		formBuilder.add(FormBodyPartBuilder.create(name, new StringBody(value, ContentType.TEXT_PLAIN)));
	}

	public void addFormData(final String name, final String[] value) {
		if (value == null) {
			return;
		}
		for (final String v: value) {
			formBuilder.add(FormBodyPartBuilder.create(name, new StringBody(v, ContentType.TEXT_PLAIN)));
		}
	}


	private static int getIntProperty(final String propertyName, final int defaultValue){
		int value=0;
		try{
			final ResourceBundle rb = ResourceBundle.getBundle(PROPERTY_FILE); // ResourceBundle caches, so no harm in calling this each time
			final String valueString = rb.getString(propertyName);
			if ((valueString==null) || (valueString.trim().isEmpty())){
				// if blank, create MissingResourceException just as if the property didn't exist
				throw new MissingResourceException("Can't find resource for bundle " + rb.getClass().getName() + ", key " + propertyName, rb.getClass().getName(), propertyName);
			}
			value = Integer.parseInt(valueString.trim());
		} catch (final MissingResourceException e){
			// Log these as errors. We really do want these to be defined in the property file, but we don't want to stop the class from working
			LOG.error("Property '" + propertyName + "' not found. Using default value '" + defaultValue +"'.  Exception message: " + e.getMessage());
			value = defaultValue;
		} catch (final NumberFormatException e){
			LOG.error("Property '" + propertyName + "' is not a valid number. Using default value '" + defaultValue +"'.  Exception message: " + e.getMessage());
			value = defaultValue;
		} catch (final Throwable t){
			LOG.error("Exception thrown while retrieving Property '" + propertyName + "'! Using default value '" + defaultValue +"'.  Exception message: " + t.getMessage());
			value = defaultValue;
		}
		return value;
	}

	private static void checkAndLogStats(){
		final PoolStats stats = cm.getTotalStats();
		final int pendingCount = stats.getPending();
		if (LOG.isDebugEnabled()){
			final StringBuffer buf = new StringBuffer("Resource Pool Stats. ");
			buf.append("Max: ").append(stats.getMax());
			buf.append(", Leased: ").append(stats.getLeased());
			buf.append(", Pending: ").append(pendingCount);
			buf.append(", Available: ").append(stats.getAvailable());
			LOG.debug(buf.toString());
		}
		if (stats.getPending()>0){
			LOG.warn("There are pending Resource Requests! Pending count: " + pendingCount);
		}
	}
}
